/*
 * Board configuration for HW-A03
 *
 * Copyright (C) 2016 SHARP.Co.
 *
 * Based on u-boot/board/rskrza1/rskrza1.c
 *
 * This file is released under the terms of GPL v2 and any later version.
 * See the file COPYING in the root directory of the source tree for details.
 */

#include <common.h>
#include <asm/io.h>
#include <asm/processor.h>
#include <i2c.h>
#include <rtc.h>
#include <version.h>
#include <asm/arch/rza1-regs.h>

#include <spi.h>
#include <spi_flash.h>

#include "updater.h"
#include "rescue.h"

/* This function temporary disables the feature of OR-ing the results of
   commands together when using dual spi flash memory. When using single
   flash, it does nothing. */
void qspi_disable_auto_combine(void);
void disable_uart_brk_pin(void);
static void wait(void);
static int led_sw_flag;
extern int serial_assign(const char *name);

//#define DEBUG

/* Port Function Registers */
#define PORTn_base 0xFCFE3000
#define Pn(n)	(PORTn_base + 0x0000 + n * 4)	/* Port register R/W */
#define PSRn(n)	(PORTn_base + 0x0100 + n * 4)	/* Port set/reset register R/W */
#define PPRn(n)	(PORTn_base + 0x0200 + n * 4)	/* Port pin read register R */
#define PMn(n)	(PORTn_base + 0x0300 + n * 4)	/* Port mode register R/W */
#define PMCn(n)	(PORTn_base + 0x0400 + n * 4)	/* Port mode control register R/W */
#define PFCn(n)	(PORTn_base + 0x0500 + n * 4)	/* Port function control register R/W */
#define PFCEn(n)	(PORTn_base + 0x0600 + n * 4)	/* Port function control expansion register R/W */
#define PNOTn(n)	(PORTn_base + 0x0700 + n * 4)	/* Port NOT register W */
#define PMSRn(n)	(PORTn_base + 0x0800 + n * 4)	/* Port mode set/reset register R/W */
#define PMCSRn(n)	(PORTn_base + 0x0900 + n * 4)	/* Port mode control set/reset register R/W */
#define PFCAEn(n)	(PORTn_base + 0x0A00 + n * 4)	/* Port Function Control Additional Expansion register R/W */
#define PIBCn(n)	(PORTn_base + 0x4000 + n * 4)	/* Port input buffer control register R/W */
#define PBDCn(n)	(PORTn_base + 0x4100 + n * 4)	/* Port bi-direction control register R/W */
#define PIPCn(n)	(PORTn_base + 0x4200 + n * 4)	/* Port IP control register R/W */

enum pfc_pin_alt_mode {ALT1=1, ALT2, ALT3, ALT4, ALT5, ALT6, ALT7, ALT8};
enum pfc_pin_gpio_mode {GPIO_OUT=0, GPIO_IN=1};

const u32 alt_settings[9][3] = {
	/* PFCAEn, PFCEn, PFCn */
	{0,0,0},/* dummy */
	{0,0,0},/* 1st alternative function */
	{0,0,1},/* 2nd alternative function */
	{0,1,0},/* 3rd alternative function */
	{0,1,1},/* 4th alternative function */
	{1,0,0},/* 5th alternative function */
	{1,0,1},/* 6th alternative function */
	{1,1,0},/* 7th alternative function */
	{1,1,1},/* 8th alternative function */
};

void pfc_set_value(u8 n, u8 b, u8 data)
{
	*(u32 *)PSRn(n) = 1UL<<(b+16) | (u32)data<< b;
}

/* Arguments:
   n = port(1-11)
   b = bit(0-15)
   d = direction('GPIO_IN','GPIO_OUT')
*/
void pfc_set_gpio(u8 n, u8 b, u8 d)
{
	*(u32 *)PMCSRn(n) = 1UL<<(b+16);		// Pin as GPIO
	*(u32 *)PMSRn(n) = 1UL<<(b+16) | (u32)d << b;	// Set pin IN/OUT

	if( d == GPIO_IN )
		*(u16 *)PIBCn(n) |= 1UL << b; 		// Enable input buffer
	else
		*(u16 *)PIBCn(n) &= ~(1UL << b); 	// Disable input buffer

    wait();
}

/* Arguments:
   n = port(1-11)
   b = bit(0-15)
   v = value (0 or 1)
*/
void gpio_set(u8 n, u8 b, u8 v)
{
	/* The pin should have been configured as GPIO_OUT using pfc_set_gpio */

	/* Use the 'Port Set and Reset Register' to only effect the pin bit */
	/* Upper WORD is the bit mask, the lower WORD is the desire pin level */
	*(u32 *)PSRn(n) = 1UL<<(b+16) | (u32)v<< b;	// Set pin to 0/1
}

/* Arguments:
   n = port(1-11)
   b = bit(0-15)
   return = current pin level (0 or 1);
*/
u8 gpio_read(u8 n, u8 b)
{
	/* The pin should have been configured as GPIO_IN using pfc_set_gpio */
	return ( *(volatile u16 *)PPRn(n) >> b ) & 0x0001;
}

/* Arguments:
    n = port number (P1-P11)
    b = bit number (0-15)
    alt = Alternative mode ('ALT1'-'ALT7')
    inbuf =  Input buffer (0=disabled, 1=enabled)
    bi = Bidirectional mode (0=disabled, 1=enabled)
*/
void pfc_set_pin_function(u16 n, u16 b, u16 alt, u16 inbuf, u16 bi)
{
	u16 value;

	/* Set PFCAEn */
	value = *(u16 *)PFCAEn(n);
	value &= ~(1UL<<b); // clear
	value |= (alt_settings[alt][0] & 1UL) << b; // set(maybe)
	*(u16 *)PFCAEn(n) = value;

	/* Set PFCEn */
	value = *(u16 *)PFCEn(n);
	value &= ~(1UL<<b); // clear
	value |= (alt_settings[alt][1] & 1UL) << b; // set(maybe)
	*(u16 *)PFCEn(n) = value;

	/* Set PFCn */
	value = *(u16 *)PFCn(n);
	value &= ~(1UL<<b); // clear
	value |= (alt_settings[alt][2] & 1UL) << b; // set(maybe)
	*(u16 *)PFCn(n) = value;

	/* Set Pn */
	/* Not used for alternative mode */
	/* NOTE: PIP must be set to '0' for the follow peripherals and Pn must be set instead
		<> Multi-function timer pulse unit
		<> LVDS output interface
		<> Serial sound interface
	   For those, use this to set Pn: *(u32 *)PSRn(n) = 1UL<<(b+16) | direction<<(b);
	*/

	/* Set PIBCn */
	value = *(u16 *)PIBCn(n);
	value &= ~(1UL<<b); // clear
	value |= inbuf << b; // set(maybe)
	*(u16 *)PIBCn(n) = value;

	/* Set PBDCn */
	value = *(u16 *)PBDCn(n);
	value &= ~(1UL<<b); // clear
	value |= bi << b; // set(maybe)
	*(u16 *)PBDCn(n) = value;

	/* Alternative mode '1' (not GPIO '0') */
	*(u32 *)PMCSRn(n) |= 1UL<<(b+16) | 1UL<<(b);

	/* Set PIPCn so pin used for function '1'(not GPIO '0') */
	*(u16 *)PIPCn(n) |= 1UL <<b;
}


int spi_flash_cmd(struct spi_slave *spi, u8 cmd, void *response, size_t len);
struct spi_flash *spi_flash_probe(unsigned int bus, unsigned int cs,
		unsigned int max_hz, unsigned int spi_mode);
int spi_flash_cmd_write(struct spi_slave *spi, const u8 *cmd, size_t cmd_len,
		const void *data, size_t data_len);

DECLARE_GLOBAL_DATA_PTR;

int checkboard(void)
{
	puts("BOARD: SHARP A03\n");
	return 0;
}

int board_init(void)
{
	gd->bd->bi_boot_params = (CONFIG_SYS_SDRAM_BASE + 0x100);
	return 0;
}

#define REPEAT_D 0x000033F1
static void wait()
{
	int i;

	for (i=0;i<REPEAT_D;i++) {
		asm("nop");
	}
}

int board_early_init_f(void)
{
	int i;

	/* This function runs early in the boot process, before u-boot is relocated
	   to RAM (hence the '_f' in the function name stands for 'still running from
	   flash'). A temporary stack has been set up for us which is why we can
	   have this as C code. */

	/* When booting from Parallel NOR, some pins need to be bi-directional */
	/* CS0, RD, A1-A15 */
	#if defined(CONFIG_BOOT_MODE0)
	  #define NOR_BIDIR 1
	  /* TODO: Replace '0' with 'NOR_BIDIR for those pins below */
	#else
	  #define NOR_BIDIR 0
	#endif

	rtc_reset();	/* to start rtc */

	/* =========== Pin Setup =========== */
	/* Specific for the RZ/H on the RSK board. Adjust for your board as needed. */

    /* Service UART (Serial Console/Service UART) */
    pfc_set_pin_function(7, 1,  ALT4, 0, 0);	/* P7_1  = TxD2 */
    pfc_set_pin_function(1, 10, ALT3, 0, 0);	/* P1_10 = RxD2 */

    /* Appliance UART */
    pfc_set_pin_function(3, 15, ALT5, 0, 0);    /* P3_15 = TxD1 */
    pfc_set_pin_function(1, 9,  ALT3, 0, 0);    /* P1_10 = RxD1 */
    pfc_set_pin_function(4, 0,  ALT5, 0, 1);    /* P4_0  = CTS1 */
    pfc_set_pin_function(4, 1,  ALT5, 0, 1);    /* P4_1  = RTS1 */

#if defined(CONFIG_A03_DEBUG)
    /* (for debug) */
    pfc_set_pin_function(6, 14, ALT5, 0, 0);   /* P6_14 = RxD0 */
    pfc_set_pin_function(6, 15, ALT5, 0, 0);   /* P6_15 = TxD0 */
    pfc_set_pin_function(6, 12, ALT5, 0, 1);   /* P6_12 = RTS0 */
    pfc_set_pin_function(6, 13, ALT5, 0, 1);   /* P6_13 = CTS0 */
#else
    pfc_set_gpio(6, 12, GPIO_OUT);
    pfc_set_gpio(6, 13, GPIO_OUT);
    pfc_set_gpio(6, 14, GPIO_OUT);
    pfc_set_gpio(6, 15, GPIO_OUT);
#endif

    /* QSPI_0 ch0 (booted in 1-bit, need to change to 4-bit) */
    pfc_set_pin_function(4, 4, ALT2, 0, 0);	/* P4_4 = SPBCLK_0 */
    pfc_set_pin_function(4, 5, ALT2, 0, 0);	/* P4_5 = SPBSSL_0 */
    pfc_set_pin_function(4, 6, ALT2, 0, 1);	/* P4_6 = SPBIO00_0 (bi dir) */
    pfc_set_pin_function(4, 7, ALT2, 0, 1);	/* P4_7 = SPBIO10_0 (bi dir) */
    pfc_set_pin_function(4, 2, ALT2, 0, 1);	/* P4_2 = SPBIO20_0 (bi dir) */
    pfc_set_pin_function(4, 3, ALT2, 0, 1);	/* P4_3 = SPBIO30_0 (bi dir) */

    /* RIIC Ch 1 */
    pfc_set_pin_function(1, 2, ALT1, 0, 1);	/* P1_2 = RIIC1SCL (bi dir) */
    pfc_set_pin_function(1, 3, ALT1, 0, 1);	/* P1_3 = RIIC1SDA (bi dir) */
    /* PLL clock */
    pfc_set_pin_function(2, 7, ALT3, 0, 1);	/* P2_7 = TIOC1A */
    /* I2S ch.0 */
    pfc_set_pin_function(6, 9, ALT3, 0, 1);	/* P6_9 = SSIWS0 */
    pfc_set_pin_function(6, 10, ALT3, 0, 0);/* P6_10 = SSITxD0 */
    pfc_set_pin_function(6, 11, ALT3, 0, 0);/* P6_11 = SSIRxD0 */
    pfc_set_pin_function(6, 8, ALT3, 0, 1);	/* P6_8 = SSISCK0 */

    /* GPIO */
    pfc_set_gpio(1, 11, GPIO_IN);   /* P1_11 = USB/OC# */

    pfc_set_gpio(6, 0, GPIO_OUT);	/* P6_0 = CODEC:RESET */

    pfc_set_gpio(6, 1, GPIO_OUT);   /* P6_1 = USB:VBUS EN */
    pfc_set_gpio(6, 5, GPIO_IN);    /* P6_5 = WLAN:IRQ (not used) */
    pfc_set_gpio(6, 4, GPIO_OUT);   /* P6_4 = WLAN:RESET */

    pfc_set_gpio(1, 5, GPIO_OUT);   /* P1_5 = IO_2 LED2 */
    pfc_set_gpio(1, 4, GPIO_IN);    /* P1_4 = IO_4 SW1 */
    pfc_set_gpio(1, 6, GPIO_IN);    /* P1_6 = IS_5 SW2 */
    pfc_set_gpio(1, 7, GPIO_OUT);   /* P1_7 = IO_3 LED3 */
    pfc_set_gpio(1, 1, GPIO_OUT);   /* P1_1 = IO_1 LED1 */
    pfc_set_gpio(1, 8, GPIO_IN);    /* P1_8 = IO_6 SW/LED */
    wait();

#if defined(CONFIG_A03_DEBUG) /* Set 1 if you use SDRAM */
    /* SDRAM */
    pfc_set_pin_function(2, 0, ALT1, 0, 0);	/* P2_0 = CS3 */
    for(i=0;i<=15;i++)
        pfc_set_pin_function(5, i, ALT1, 0, 1);	/* P5_0~15 = D0-D15 (bi dir) */
    pfc_set_pin_function(2, 1, ALT1, 0, 0);	/* P2_1 = RAS */
    pfc_set_pin_function(2, 2, ALT1, 0, 0);	/* P2_2 = CAS */
    pfc_set_pin_function(2, 3, ALT1, 0, 0);	/* P2_3 = CKE */
    pfc_set_pin_function(2, 6, ALT1, 0, 0);	/* P2_6 = RD/WR */
    pfc_set_pin_function(2, 4, ALT1, 0, 0);	/* P2_4 = WE0/DQMLL */
    pfc_set_pin_function(2, 5, ALT1, 0, 0);	/* P2_5 = WE1/DQMLU */
    for(i=0;i<=14;i++)
        pfc_set_pin_function(3, i, ALT1, 0, 0);	/* P3_0~15 = A1-A15 (bi dir) */

    /**********************************************/
    /* Configure SDRAM (CS2, CS3)                 */
    /**********************************************/
    /* If SDRAM is connected at only one area, CS3 must be configured as SDRAM.
       In that case CS2 must be configured as a normal or a byte selection SRAM.
    */
    #define CS2BCR_D	0x00000C00  // Normal area, 16bit
    #define CS2WCR_D	0x00000000
    #define CS3BCR_D	0x00004C00  // SDRAM, 16bit
    #define CS3WCR_D	0x0000649B
    #define SDCR_D		0x00000811
    #define RTCOR_D		0xa55a0020
    #define RTCSR_D		0xa55a0010
    *(u32 *)CS2BCR = CS2BCR_D;
    *(u32 *)CS2WCR = CS2WCR_D;
    *(u32 *)CS3BCR = CS3BCR_D;
    *(u32 *)CS3WCR = CS3WCR_D;
    *(u32 *)SDCR = SDCR_D;
    *(u32 *)RTCOR = RTCOR_D;
    *(u32 *)RTCSR = RTCSR_D;

    /* wait */
    wait();
    /* [power on sequence]
        The final step is to set the SDRAM Mode Register by written to a
        specific address (the data value is ignored) */
    //#define SDRAM_MODE_CS2 0x3FFFD040
    //*(u32 *)SDRAM_MODE_CS2 = 0;
    #define SDRAM_MODE_CS3 0x3FFFE040
    *(u32 *)SDRAM_MODE_CS3 = 0;

#else /* care of unused ports */
	/* Port 1 */
	pfc_set_gpio(1, 0, GPIO_OUT);
	for(i=12; i<=15; i++)
		pfc_set_gpio(1, i, GPIO_OUT);

	/* Port 2 */
	for(i=0; i<=6; i++)
		pfc_set_gpio(2, i, GPIO_OUT);
	pfc_set_gpio(2, 9, GPIO_OUT);

	/* Port 3 */
	for(i=0; i<=14; i++)
		pfc_set_gpio(3, i, GPIO_OUT);

	/* Port 5 */
	for(i=0; i<=15; i++)
		pfc_set_gpio(5, i, GPIO_OUT);

	/* Port 6 */
	pfc_set_gpio(6, 2, GPIO_OUT);
	pfc_set_gpio(6, 3, GPIO_OUT);
	pfc_set_gpio(6, 6, GPIO_OUT);

	/* Port 7 */
	pfc_set_gpio(7, 0, GPIO_OUT);
	for(i=8; i<=11; i++)
		pfc_set_gpio(7, i, GPIO_OUT);

	/* Port 8 */
	for(i=0; i<=15; i++)
		pfc_set_gpio(8, i, GPIO_OUT);

	/* Port 9 */
	for(i=0; i<=5; i++)
		pfc_set_gpio(9, i, GPIO_OUT);

#endif

    return 0;
}

void start_a03(int force)
{
	int ret = 0;

	START_MODE mode = NORMAL_MODE;
	ulong (*entry)(void);

	/* BRK信号送出の為にGPIOにPIN設定を変更する */
	/*   UARTのTxD1の変更は kernelにジャンプする直前に行う */
	enable_uart_brk_pin();

	if (led_sw_board() == (u8)UPDATER_LED_SW_CONNECTED) {
		led_sw_flag = 1;
	}
	printf("%s LED/SW Board\n", 
		(led_sw_flag == 1) ? "Detected": "Not detected");

	mode = get_startmode();

	if (force & FORCE_UPDATER) {
		mode = UPDATER_MODE;
		printf("force updater mode\n");
	} 

	if (force & FORCE_RESCUE) {
		mode = RESCUE_MODE;
		printf("force rescue mode\n");
		ret = updater_initial_dct();
		if (ret != 0) {
			mode = RESET_MODE;
		}
	}

	if (force & FORCE_HALT) {
		mode =RESCUE_HALT_MODE;
		printf("force halt mode\n");
	}

	DBG_PRINT("bootstate:  %d\n", mode);

	if (mode == RESET_MODE) {
		/* リセットを実施 */
		printf("initial DCT read error reset ...\n");
		reset_cpu(0);
	}

	if((mode == RESCUE_HALT_MODE) || (mode == RESCUE_MODE)) {
 		/* レスキューモード遷移後は フラグを落とす */
		set_startmode(RESCUE_FLAG_RESET);

		/* レスキュー起動 */
		serial_assign("sh_serial2");
		printf("%s mode ....\n", (mode == RESCUE_HALT_MODE)?"Halt":"Rescue");
		ret = load_rescue((u32)UPDATER_LOAD_ADDR);
		wlan_led(LED_OFF);
		link_led(LED_OFF);
		power_led(LED_OFF);
		if (ret == 0) {
			printf("rescue: load success\n");
			entry = (ulong (*)(void))UPDATER_LOAD_ADDR;
			entry(); /* ジャンプ後は戻ってこない */
		} else {
			printf("rescue: load error\n");
			/* リセットを実施 */
			reset_cpu(0);
		}
	} else if (mode == UPDATER_MODE) {
		/* アップデータ起動 */
		printf("updater mode ....\n");
		ret = load_updater((u32)UPDATER_LOAD_ADDR);
		if(ret == 0) {
			/* アップデータ成功の場合はアップデータへジャンプする */
			printf("updater: load success\n");
			entry = (ulong (*)(void))UPDATER_LOAD_ADDR;
			entry(); /* ジャンプ後は戻ってこない */
		} else {
			printf("updater: load error\n");

 			/* アップデータ起動失敗時はフラグを落とす */
			set_startmode(UPDATE_FLAG_RESET);

			/* リセットを実施 */
			reset_cpu(0);
		}
	}
}

void make_a03_version (char *string, u32 version)
{
	int x, y, z;

	if (string == NULL)
		return;

	x =  (version & 0x000000ff);
	y =  (version & 0x0000ff00) >> 8;
	z = ((version & 0x00ff0000) >> 8) |
		((version & 0xff000000) >> 24);

	if (x) {
		/* Release Version */
		sprintf(string, "%d.%d.%d", x, y, z);
	} else {
		/* Debug Version */
		sprintf(string, "%d.%02d.%05d", x, y, z);
	}
}

int board_late_init(void)
{
   char version_string [32];

    /* Reset of audio codec */
    pfc_set_value(6, 0, 1);
    wait();
    pfc_set_value(6, 0, 0);
    wait();
    pfc_set_value(6, 0, 1);
    wait();

    /* Initialize LED (high is off) */
    pfc_set_value(1, 1, 1);
    wait();
    pfc_set_value(1, 5, 1);
    wait();
    pfc_set_value(1, 7, 1);
    wait();

    i2c_init(CONFIG_SYS_I2C_SPEED, 0);
    i2c_set_bus_num(CONFIG_SYS_I2C_MODULE);

#if 0
	printf(	"\t\t      SPI Flash Memory Map\n"
		"\t\t------------------------------------\n"
		"\t\t         Start      Size     SPI\n");
	printf(	"\t\tu-boot:  0x%08X 0x%06X 0\n", 0,CONFIG_ENV_OFFSET);
	printf(	"\t\t   env:  0x%08X 0x%06X 0\n", CONFIG_ENV_OFFSET, CONFIG_ENV_SIZE);
	printf(	"\t\t    DT:  0x%08X 0x%06X 0\n", DTD_PTN_OFFSET,DTD_PTN_SIZE);
	printf(	"\t\tKernel:  0x%08X 0x%06X 0+1 (size*=2)\n",KERNEL_PTN_OFFSET, KERNEL_PTN_SIZE);
	printf(	"\t\trootfs:  0x%08X 0x%06X 0+1 (size*=2)\n",ROOTFS_PTN_OFFSET, ROOTFS_PTN_SIZE);
#endif

	/* Boot uImage in external SDRAM */
	/* Rootfs is a squashfs image in memory mapped QSPI */
	/* => run s_boot */
	setenv("s1", "sf probe 0; sf read 09800000 C0000 8000"); // Read out DT blob
	setenv("s2", "sf probe 0:1; sf read 09000000 100000 500000"); //Copy Kernel to SDRAM
	setenv("s3", "bootm start 0x09000000 - 0x09800000 ; bootm loados ;"\
			"fdt memory 0x08000000 0x02000000"); // Change memory address in DTB
	setenv("s4", "qspi dual"); // Change XIP interface to dual QSPI
	setenv("sargs", "console=ttySC2,115200 console=tty0 ignore_loglevel root=/dev/mtdblock0"); // bootargs
	setenv("s_boot", "run s1 s2 s3 s4; set bootargs ${sargs} uboot_version=${a03_version}; fdt chosen; bootm go"); // run the commands

	/* Boot XIP using internal RAM */
	/* Rootfs is a squashfs image in memory mapped QSPI */
	/* => run x_boot */
	/* Read out DT blob */
	setenv("x1", "sf probe 0; sf read 20500000 C0000 8000");
	/* Change memory address in DTB */
	setenv("x2", "fdt addr 20500000 ; fdt memory 0x20000000 0x00A00000"); /* 10MB RAM */
	/* Change XIP interface to dual QSPI */
	setenv("x3", "qspi dual");
	setenv("xargs", "console=ttySC2,115200 console=tty0 ignore_loglevel root=/dev/mtdblock0"); // bootargs
	setenv("x_boot", "run x1 x2 x3; set bootargs ${xargs} uboot_version=${a03_version}; fdt chosen; bootx 18200000 20500000"); // run the commands

	/* Boot XIP using internal RAM */
	/* Rootfs is a AXFS image in memory mapped QSPI */
	/* => run xa_boot */
	/* Read out DT blob */
	setenv("xa1", "sf probe 0; sf read 202FC000 C0000 4000");
	/* Change memory address in DTB */
	setenv("xa2", "fdt addr 202FC000 ; fdt memory 0x20000000 0x00300000"); /* 3MB RAM */
	/* Change XIP interface to dual QSPI */
	setenv("xa3", "qspi single");
	setenv("xaargs", "console=ttySC0,115200 ignore_loglevel rz_irq_trim rootflags=physaddr=0x18600000 init=/root/start.sh"); // bootargs
	setenv("xa_boot", "run xa1 xa2 xa3; set bootargs ${xaargs} uboot_version=${a03_version}; fdt chosen; bootx 18200000 202FC000"); // run the commands

	/* Boot XIP using external SDRAM RAM */
	/* Rootfs is a AXFS image in memory mapped QSPI */
	/* => run xsa_boot */
	/* Read out DT blob */
	setenv("xsa1", "sf probe 0; sf read 0D800000 C0000 8000");
	/* Change memory address in DTB */
	setenv("xsa2", "fdt addr 0D800000 ; fdt memory 0x0C000000 0x02000000"); /* 32MB SDRAM RAM */
	/* Change XIP interface to dual QSPI */
	setenv("xsa3", "qspi single");
	setenv("xsaargs", "console=ttySC2,115200 ignore_loglevel root=/dev/null rootflags=physaddr=0x18600000 init=/root/start.sh"); // bootargs
	setenv("xsa_boot", "run xsa1 xsa2 xsa3; set bootargs ${xsaargs} uboot_version=${a03_version}; fdt chosen; bootx 18200000 0D800000"); // run the commands

	/* Boot XIP using internal RAM (special use) */
	/* Rootfs is a AXFS image in memory mapped QSPI */
	/* => run xsh_boot */
	/* Read out DT blob */
	setenv("xsh1", "sf probe 0; sf read 202FC000 C0000 4000");
	/* Change memory address in DTB */
	setenv("xsh2", "fdt addr 202FC000 ; fdt memory 0x20000000 0x00300000"); /* 3MB RAM */
	/* Change XIP interface to dual QSPI */
	setenv("xsh3", "qspi single");
	setenv("xshargs", "console=ttySC2,115200 ignore_loglevel rz_irq_trim rootflags=physaddr=0x18600000 init=/bin/sh"); // bootargs
	setenv("xsh_boot", "run xsh1 xsh2 xsh3; set bootargs ${xshargs} uboot_version=${a03_version}; fdt chosen; bootx 18200000 202FC000"); // run the commands

	/* Boot XIP using internal RAM (dual boot) */
	/* Rootfs is a AXFS image in memory mapped QSPI */
	/* => run wl_boot */
	/* Read out DT blob */
	setenv("wl1", "sf probe 0; sf read 202FC000 C0000 4000");
	/* Change memory address in DTB */
	setenv("wl2", "fdt addr 202FC000 ; fdt memory 0x20000000 0x00300000"); /* 3MB RAM */
	/* Change XIP interface to dual QSPI */
	setenv("wl3", "qspi single");
	setenv("wlargs", "console=ttySC2,115200 ignore_loglevel rz_irq_trim rootflags=physaddr=0x19600000 init=/root/start.sh"); // bootargs
	setenv("wl_boot", "run wl1 wl2 wl3; set bootargs ${wlargs} uboot_version=${a03_version}; fdt chosen; bootx 19200000 202FC000"); // run the commands

	/* Erase some block */
	setenv("clean", "sf probe; sf erase 0x00100000 +0x1000; sf erase 0x01600000 +0x1000; sf erase 0x01E00000 0x40000");

	/* u-boot version */
	make_a03_version(version_string, A03_U_BOOT_VERSION);
	printf("\nVersion: %s\n", version_string);
	setenv("a03_version", version_string);

	/* Run linux boot */
	setenv("bootcmd", "run xa_boot");

/* アップデータまたはレスキューの起動処理 */
	start_a03(0);

	return 0;
}

int dram_init(void)
{
#if (1 !=  CONFIG_NR_DRAM_BANKS)
# error CONFIG_NR_DRAM_BANKS must set 1 in this board.
#endif
	gd->bd->bi_dram[0].start = CONFIG_SYS_SDRAM_BASE;
	gd->bd->bi_dram[0].size = CONFIG_SYS_SDRAM_SIZE;
	gd->ram_size = CONFIG_SYS_SDRAM_SIZE * CONFIG_NR_DRAM_BANKS;

	return 0;
}

#define	UPDATER_WDT_CMASK	(0xA500)
#define	UPDATER_WDT_DMASK	(0x5A00)
void reset_cpu(ulong addr)
{
	/* WDT を有効にしてリセットする */
	*(volatile u16 *)WTCSR = UPDATER_WDT_CMASK | ((*(volatile u16 *)WTCSR & (~(1<<5))) & 0x00FF);		// TME = 0 タイマーディスエーブル
	*(volatile u16 *)WTCSR = UPDATER_WDT_CMASK | ((*(volatile u16 *)WTCSR | (1<<6)) & 0xFF);			// WT/IT = 1 ウォッチドッグモード
	*(volatile u16 *)WTCSR = UPDATER_WDT_CMASK | (*(volatile u16 *)WTCSR & 0x00FC);					// CKS[2:0] = 000 クロック分周比 1
	*(volatile u16 *)WRCSR = UPDATER_WDT_DMASK | ((*(volatile u16 *)WRCSR | (1<<6))&0x00FF);		// RSTE = 1 リセットイネーブル
	*(volatile u16 *)WRCSR = UPDATER_WDT_CMASK | ((*(volatile u16 *)WRCSR & (~(1<<7)))&0x00FF);		// WOVF = 0 カウンタクリア
	*(volatile u16 *)WTCNT = UPDATER_WDT_DMASK | 0x00FF;									// タイマ・カウンタ初期値  0xFE
	*(volatile u16 *)WTCSR = UPDATER_WDT_CMASK | ((*(volatile u16 *)WTCSR | (1<<5))&0x00FF);		// TME = 1 タイマーイネーブル

	for(;;){}
	return;
}

void led_set_state(unsigned short value)
{
}

/* XIP Kernel boot */
int do_bootx(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	ulong machid = MACH_TYPE_RSKRZA1;
	void (*kernel_entry)(int zero, int arch, uint params);
	ulong r2;
	ulong img_addr;
	char *endp;

	/* need at least two arguments */
	if (argc < 2)
		goto usage;

	img_addr = simple_strtoul(argv[1], &endp, 16);
	kernel_entry = (void (*)(int, int, uint))img_addr;

#ifdef CONFIG_USB_DEVICE
	udc_disconnect();
#endif
	cleanup_before_linux();

	r2 = simple_strtoul(argv[2], NULL, 16);

	/* The kernel expects the following when booting:
	 *  r0 - 0
	 *  r1 - machine type
	 *  r2 - boot data (atags/dt) pointer
	 *
	 * For more info, refer to:
	 *  https://www.kernel.org/doc/Documentation/arm/Booting
	 */

	printf("Booting Linux...\n");

	/* UART TxD1に変更 */
	disable_uart_brk_pin();

	kernel_entry(0, machid, r2);

	return 0;

usage:
	return CMD_RET_USAGE;
}
static char bootx_help_text[] =
	"x_addr dt_addr\n    - boot XIP kernel in Flash\n"
	"\t x_addr: Address of XIP kernel in Flash\n"
	"\tdt_addr: Address of Device Tree blob image";
U_BOOT_CMD(
	bootx,	CONFIG_SYS_MAXARGS,	1,	do_bootx,
	"boot XIP kernel in Flash", bootx_help_text
);


#define CMNCR_0	0x3FEFA000	/* Common control register */
#define DRCR_0	0x3FEFA00C	/* Data Read Control Register */
#define DRCMR_0	0x3FEFA010	/* Data Read Command Setting Register */
#define DREAR_0 0x3FEFA014	/* Data read extended address setting register */
#define DRENR_0 0x3FEFA01C	/* Data read enable setting register */
#define DROPR_0 0x3FEFA018	/* Data read option setting register */
#define DMDMCR_0 0x3FEFA058	/* SPI Mode Dummy Cycle Setting Register */
#define DRDRENR_0 0x3FEFA05C	/* Data Read DDR Enable Register */


struct read_mode {
	u8 cmd;
	char name[50];
};
#define READ_MODES 9
const struct read_mode modes[READ_MODES] = {
	{0x03, "Read Mode (3-byte Addr) (RZ/A1 reset value)"},
	{0x0C, "Fast Read Mode (4-byte Addr)"},
	{0x6C, "Quad Read Mode (4-byte Addr)"},
	{0xEC, "Quad I/O Read Mode (4-byte Addr)"},
	{0xEE, "Quad I/O DDR Read Mode (4-byte Addr)"},
	{0x0B, "Fast Read Mode (3-byte Addr)"},
	{0x6B, "Quad Read Mode (3-byte Addr)"},
	{0xEB, "Quad I/O Read Mode (3-byte Addr)"},
	{0xED, "Quad I/O DDR Read Mode (3-byte Addr)"},
};

/* If you are using a SPI Flash device that does not have 4-byte address
   commands (Flash size <= 16MB), then change the #if 0 to #if 1 */
#if 0
 #define ADDRESS_BYTE_SIZE 3	/* Addresses are 3-bytes (A0-A23) */
 #define FAST_READ 0x0B		/* Fast Read Mode (1-bit cmd, 1-bit addr, 1-bit data, 3-bytes of address) */
 #define QUAD_READ 0x6B		/* Quad Read Mode (1-bit cmd, 1-bit addr, 4-bit data, 3-bytes of address) */
 #define QUAD_IO_READ 0xEB	/* Quad I/O Read Mode (1-bit cmd, 4-bit addr, 4-bit data, 3-bytes of address) */
 #define QUAD_IO_DDR_READ 0xED	/* Quad I/O DDR Read Mode (1-bit cmd, 1-bit addr, 4-bit data, 3-bytes of address) */
#else
 #define ADDRESS_BYTE_SIZE 4	/* Addresses are 4-bytes (A0-A31) */
 #define FAST_READ 0x0C		/* Fast Read Mode (1-bit cmd, 1-bit addr, 1-bit data, 4-bytes of address) */
 #define QUAD_READ 0x6C		/* Quad Read Mode (1-bit cmd, 1-bit addr, 4-bit data, 4-bytes of address) */
 #define QUAD_IO_READ 0xEC	/* Quad I/O Read Mode (1-bit cmd, 4-bit addr, 4-bit data, 4-bytes of address) */
 #define QUAD_IO_DDR_READ 0xEE	/* Quad I/O DDR Read Mode (1-bit cmd, 1-bit addr, 4-bit data, 4-bytes of address) */
#endif

/* These should be filled out for each device */
u32 g_FAST_RD_DMY;		/* Fast Read Mode */
u32 g_QUAD_RD_DMY;		/* Quad Read Mode  */
u32 g_QUAD_IO_RD_DMY;		/* Quad I/O Read Mode  */
u32 g_QUAD_IO_DDR_RD_DMY;	/* Quad I/O DDR Read Mode  */
u32 g_QUAD_IO_RD_OPT;		/* Addtional option or 'mode' settings */

/**********************/
/* Spansion S25FL512S */
/**********************/
int enable_quad_spansion(struct spi_flash *sf, u8 quad_addr, u8 quad_data )
{
	/* NOTE: Macronix and Windbond are similar to Spansion */
	/* NOTE: Once quad comamnds are enabled, you don't need to disable
		 them to use the non-quad mode commands, so we just always
		 leave them on. */
	int ret = 0;
	u8 data[5];
	u8 cmd;
	u8 spi_cnt = 1;
	u8 st_reg[2];
	u8 cfg_reg[2];

	/* Read ID Register (for cases where not all parts are the same) */
	//ret |= spi_flash_cmd(sf->spi, 0x9F, &data[0], 5);

	if (sf->spi->cs)
		spi_cnt = 2; /* Dual SPI Flash */

	/* Read Status register (RDSR1 05h) */
	qspi_disable_auto_combine();
	ret |= spi_flash_cmd(sf->spi, 0x05, st_reg, 1*spi_cnt);

	/* Read Configuration register (RDCR 35h) */
	qspi_disable_auto_combine();
	ret |= spi_flash_cmd(sf->spi, 0x35, cfg_reg, 1*spi_cnt);

#ifdef DEBUG
	printf("Initial Values:\n");
	for(cmd = 0; cmd <= spi_cnt; cmd++) {
		printf("   SPI Flash %d:\n", cmd+1);
		printf("\tStatus register = %02X\n", st_reg[cmd]);
		printf("\tConfiguration register = %02X\n", cfg_reg[cmd]);
	}
#endif

	/* Skip SPI Flash configure if already correct */
	/* Note that if dual SPI flash, both have to be set */
	if ( (cfg_reg[0] != 0x02 ) ||
	     ((spi_cnt == 2) && (cfg_reg[1] != 0x02 ))) {

		data[0] = 0x00;	/* status reg: Don't Care */
		data[1] = 0x02; /* confg reg: Set QUAD, LC=00b */

		/* Send Write Enable (WREN 06h) */
		ret |= spi_flash_cmd(sf->spi, 0x06, NULL, 0);

		/* Send Write Registers (WRR 01h) */
		cmd = 0x01;
		ret |= spi_flash_cmd_write(sf->spi, &cmd, 1, data, 2);

		/* Wait till WIP clears */
		do
			spi_flash_cmd(sf->spi, 0x05, &data[0], 1);
		while( data[0] & 0x01 );

	}

#ifdef DEBUG
	/* Read Status register (RDSR1 05h) */
	qspi_disable_auto_combine();
	ret |= spi_flash_cmd(sf->spi, 0x05, st_reg, 1*spi_cnt);

	/* Read Configuration register (RDCR 35h) */
	qspi_disable_auto_combine();
	ret |= spi_flash_cmd(sf->spi, 0x35, cfg_reg, 1*spi_cnt);

	printf("New Values:\n");
	for(cmd = 0; cmd <= spi_cnt; cmd++) {
		printf("   SPI Flash %d:\n", cmd+1);
		printf("\tStatus register = %02X\n", st_reg[cmd]);
		printf("\tConfiguration register = %02X\n", cfg_reg[cmd]);
	}
#endif

	/* Finally, fill out the global settings for
	   Number of Dummy cycles between Address and Data */

	/* Spansion S25FL512S */
	/* According to the Spansion spec (Table 8.5), dummy cycles
	   are needed when LC=00 (chip default) for FAST READ,
	   QUAD READ, and QUAD I/O READ commands */
	g_FAST_RD_DMY = 8;		/* Fast Read Mode: 8 cycles */
	g_QUAD_RD_DMY = 8;		/* Quad Read Mode  */
	g_QUAD_IO_RD_DMY = 4;		/* Quad I/O Read Mode: 4 cycles */
	g_QUAD_IO_DDR_RD_DMY = 6;	/* Quad I/O DDR Read Mode  (NOT SUPPORTED) */

	/* When sending a QUAD I/O READ command, and extra MODE field
	   is needed.
	     [[ Single Data Rate, Quad I/O Read, Latency Code=00b ]]
		<> command = 1-bit, 8 clocks
		<> Addr(32bit) = 4-bit, 8 clocks,
		<> Mode = 4-bit, 2 clocks
		<> Dummy = 4-bit, 4 clocks
		<> Data = 4-bit, 2 clocks x {length}
	    See "Figure 10.37 Quad I/O Read Command Sequence" in Spansion spec
	*/
	/* Use Option data regsiters to output '0' as the
	   'Mode' field by sending OPD3 (at 4-bit) between address
	   and dummy */
	g_QUAD_IO_RD_OPT = 1;

	return ret;
}

/*******************/
/* Micron N25Q512A */
/*******************/
int enable_quad_micron(struct spi_flash *sf, u8 quad_addr, u8 quad_data )
{
/* NOTES:
	To use the QUAD commands, you need to enable dummy cycles for
	every type of FAST_READ command. While this is fine when the RZ-QSPI
	is running in XIP mode, but when you switch back to SPI mode to use
	something like "sf probe", it can't deal with those dummy cycles,
	therefore, we need to remove the dummy cycles during each
	"sf probe". See function qspi_reset_device().
	It should be noted that if the RZ/A1 is rebooted in XIP mode
	with those dummy cycles still enabled in the SPI flash, the reboot
	will still work because the RZ/A1 uses the legacy READ (0x03) command
	on reset, not FAST_READ */

	int ret = 0;
	u8 cmd;
	u8 vcfg_reg[2];

#ifdef DEBUG
	u8 st_reg[2];
	u16 nvcfg_reg[2];
	u8 tmp;

	/* Dual SPI Flash */
	if (sf->spi->cs) {
		/* Read Flag Status register (70h) */
		qspi_disable_auto_combine();	/* only lasts 1 call */
		ret |= spi_flash_cmd(sf->spi, 0x70, st_reg, 1*2);

		/* Read NONVOLATILE Configuration register (B5h) */
		qspi_disable_auto_combine();	/* only lasts 1 call */
		ret |= spi_flash_cmd(sf->spi, 0xB5, nvcfg_reg, 2*2);

		/* swap 2nd and 3rd bytes...becase data for each
		   SPI flash comes in interleaved */
		tmp = ((u8 *)nvcfg_reg)[1];
		((u8 *)nvcfg_reg)[1] = ((u8 *)nvcfg_reg)[2];
		((u8 *)nvcfg_reg)[2] = tmp;

		/* Read VOLATILE Configuration register (85h) */
		qspi_disable_auto_combine();	/* only lasts 1 call */
		ret |= spi_flash_cmd(sf->spi, 0x85, vcfg_reg, 1*2);

	}
	else {
		/* Read Flag Status register (70h) */
		ret |= spi_flash_cmd(sf->spi, 0x70, st_reg, 1);

		/* Read NONVOLATILE Configuration register (B5h) */
		ret |= spi_flash_cmd(sf->spi, 0xB5, nvcfg_reg, 2);

		/* Read VOLATILE Configuration register (85h) */
		ret |= spi_flash_cmd(sf->spi, 0x85, vcfg_reg, 1);
	}

	printf("Initial Values:\n");
	for(tmp = 0; tmp <= sf->spi->cs ;tmp++) {
		printf("   SPI Flash %d:\n", tmp+1);
		printf("\tStatus register = %02X\n", st_reg[tmp]);
		printf("\tNonVolatile Configuration register = %04X\n", nvcfg_reg[tmp]);
		printf("\tVolatile Configuration register = %02X\n", vcfg_reg[tmp]);
	}
#endif

	/* To use the QUAD commands, we need dummy cycles after every
	   FAST_READ and FAST_READ_xxx commands */
	/* Send WRITE VOLATILE CONFIGURATION REGISTER (81h) */
	cmd = 0x81;
	if( quad_addr )
		vcfg_reg[0] = 0x5B;	/* Quad IO: 5 dummy cycles */
	else if( quad_data )
		vcfg_reg[0] = 0x3B;	/* Quad Read: 3 dummy cycles */
	else
		vcfg_reg[0] = 0x0B;	/* FAST_READ: 0 dummy cycles */

	ret |= spi_flash_cmd(sf->spi, 0x06, NULL, 0);	/* Send Write Enable (06h) */
	ret |= spi_flash_cmd_write(sf->spi, &cmd, 1, vcfg_reg, 1); /* send same to both flash */

#ifdef DEBUG
	ret |= spi_flash_cmd(sf->spi, 0x70, st_reg, 1);
	ret |= spi_flash_cmd(sf->spi, 0xB5, nvcfg_reg, 2);
	ret |= spi_flash_cmd(sf->spi, 0x85, vcfg_reg, 1);
	printf("New Values:\n");
	for(tmp = 0; tmp<1;tmp++) {
		printf("   SPI Flash %d:\n", tmp+1);
		printf("\tStatus register = %02X\n", st_reg[tmp]);
		printf("\tNonVolatile Configuration register = %04X\n", nvcfg_reg[tmp]);
		printf("\tVolatile Configuration register = %02X\n", vcfg_reg[tmp]);
	}
#endif

	/* Finally, fill out the global settings for
	   Number of Dummy cycles between Address and data */

	/* Assuming a 66MHz clock. Table 13 of n25q_512mb spec */
	g_FAST_RD_DMY = 0;		/* Fast Read Mode: 0 cycles */
	g_QUAD_RD_DMY = 3;		/* Quad Read Mode: 3 cycles  */
	g_QUAD_IO_RD_DMY = 5;		/* Quad I/O Read Mode: 5 cycles  */

	g_QUAD_IO_RD_OPT = 0;		/* Quad I/O Read Mode: no additonal cycles */

	/* NOTE: This part can not run DDR at 66MHz */
	g_QUAD_IO_DDR_RD_DMY = 0;	/* Quad I/O DDR Read Mode  */

	/* HACK! */
	if( quad_addr )
	{
		/* When in QUAD I/O mode, sometimes the data is not correct.
		   It appears like the address gets corrupted. Therefore
		   we need to slow down the SPI clock in this mode. */
		/* This might be becase the board this code was devleopted on
		   had lots of wire leads attached to the SPI flash pins */
		#define	QSPI_SPBCR (0x0008)
		*(u32 *)(CONFIG_RZA1_BASE_QSPI0 + QSPI_SPBCR) = 0x0300; /* 22.22 Mbps */
		printf("\nINFO: clock is now 22.22Mbps (see function %s)\n\n",__func__);
	}

	return ret;
}

/* Dummy cycles are need for the quad mode FAST_READ commands,
   but they get applied to ever type of FAST_READ command.
   Since the "sf" commands in u-boot know nothing about dummy
   cycles, we need to shut them off if we see a "sf probe" */
int remove_dummy_micron(struct spi_flash *sf)
{
	int ret;
	u8 cmd;
	u8 cfg_reg;

#ifdef DEBUG
	/* Read VOLATILE Configuration register (85h) */
	ret = spi_flash_cmd(sf->spi, 0x85, &cfg_reg, 1);
	printf("%s: Initial Volatile Configuration register = %02X\n", __func__, cfg_reg);
#endif

	/* Send Write Enable (06h) */
	ret = spi_flash_cmd(sf->spi, 0x06, NULL, 0);

	/* Set Volatile Configuration Register to default value */
	/* Send WRITE VOLATILE CONFIGURATION REGISTER (81h) */
	cmd = 0x81;
	cfg_reg = 0xFB;
	ret |= spi_flash_cmd_write(sf->spi, &cmd, 1, &cfg_reg, 1);

#ifdef DEBUG
	/* Read Volatile Configuration register (85h) */
	ret = spi_flash_cmd(sf->spi, 0x85, &cfg_reg, 1);
	printf("%s: New Volatile Configuration register = %02X\n", __func__, cfg_reg);
#endif

	return ret;
}

/* This function is called when "sf probe" is issued, meaning that
   the user wants to access the deivce in normal single wire SPI mode.
   Since some SPI devices have to have special setups to enable QSPI mode
   or DDR QSPI mode, this function is used to reset those settings
   back to normal single wire FAST_READ mode. */
int qspi_reset_device(struct spi_flash *sf)
{
	int ret = 0;

	if( !strcmp(sf->name, "S25FL256S_256K") ) {
		/* Don't really need to do anything */
	}
	else if( !strcmp(sf->name, "S25FL256S_64K") ) {
		/* Don't really need to do anything */
	}
	else if( !strcmp(sf->name, "N25Q512") ) {
		ret = remove_dummy_micron(sf);
	}
	else {
		printf("\tWARNING: SPI Flash needs to be added to function %s()\n",__func__);
		return 1;
	}
	return ret;
}

/* QUAD SPI MODE */
int do_qspi(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	struct spi_flash *my_spi_flash;
	int ret = 0;
	int i;
	u8 cmd;
	u8 dual_chip;
	u8 quad_data;
	u8 quad_addr;
	u8 ddr;
	u32 dmdmcr, drenr, cmncr, drcmr, dropr, drdrenr;

	/* need at least 1 argument (single/dual) */
	if (argc < 2)
		goto usage;

	if ( strcmp(argv[1], "single") == 0)
		dual_chip = 0;
	else if ( strcmp(argv[1], "dual") == 0)
		dual_chip = 1;
	else
		goto usage;

	if ( argc <= 2 )
		quad_addr = 1;
	else if ( strcmp(argv[2], "a1") == 0)
		quad_addr = 0;
	else if ( strcmp(argv[2], "a4") == 0)
		quad_addr = 1;
	else
		goto usage;

	if ( argc <= 3 )
		quad_data = 1;
	else if ( strcmp(argv[3], "d1") == 0)
		quad_data = 0;
	else if ( strcmp(argv[3], "d4") == 0)
		quad_data = 1;
	else
		goto usage;

	if ( argc <= 4 )
		ddr = 0;
	else if ( strcmp(argv[4], "sdr") == 0)
		ddr = 0;
	else if ( strcmp(argv[4], "ddr") == 0)
		ddr = 1;
	else
		goto usage;

	/* checks */
	if( quad_addr && !quad_data )
		return CMD_RET_USAGE;
	if( ddr && !quad_addr )
		return CMD_RET_USAGE;

	/* Read initial register values */
	dmdmcr = *(volatile u32 *)DMDMCR_0;
	drenr = *(volatile u32 *)DRENR_0;
	cmncr = *(volatile u32 *)CMNCR_0;
	drcmr = *(volatile u32 *)DRCMR_0;
	dropr = *(volatile u32 *)DROPR_0;
	drdrenr = *(volatile u32 *)DRDRENR_0;

	printf("Current Mode: ");
	cmd = (drcmr >> 16) & 0xFF;
	for( i=0; i < READ_MODES; i++) {
		if( modes[i].cmd == cmd )
			printf("%s\n",modes[i].name);
	}

	/* bus=0, cs=0, speed=1000000 */
	if( dual_chip )
		my_spi_flash = spi_flash_probe(0, 1, 1000000, SPI_MODE_3);
	else
		my_spi_flash = spi_flash_probe(0, 0, 1000000, SPI_MODE_3);

	if (!my_spi_flash) {
		printf("Failed to initialize SPI flash.");
		return 1;
	}

	/* For Quad Mode operation, extra setup is needed in the SPI
	   Flash devices */
	if( !strcmp(my_spi_flash->name, "S25FL256S_256K") )
		ret = enable_quad_spansion(my_spi_flash, quad_addr, quad_data);
	else if( !strcmp(my_spi_flash->name, "S25FL256S_64K") )
		ret = enable_quad_spansion(my_spi_flash, quad_addr, quad_data);
	else if( !strcmp(my_spi_flash->name, "N25Q512") )
		ret = enable_quad_micron(my_spi_flash, quad_addr, quad_data);
	else
	{
		printf("ERROR: SPI Flash support needs to be added to function %s()\n",__func__);
		spi_flash_free(my_spi_flash);	/* Done with SPI Flash */
		return 1;
	}

	/* Done with SPI Flash */
	spi_flash_free(my_spi_flash);

	if ( ret )
	{
		printf("Failed to set SPI Flash Configuration register.\n");
		return 1;
	}

	/***************************/
	/* Set up RZ SPI Registers */
	/***************************/
	/* Enable data swap (SFDE) */
	/* Keeps the endian order of bytes the same on the internal bus
	   regardless of how you fetched them over SPI */
	cmncr |= 0x01000000UL;

	if( dual_chip ) {
		/* Switch to dual memory */
		cmncr |= 0x00000001UL;
	}
	else {
		/* Switch to single memory */
		cmncr &= ~0x00000001UL;
	}

	/* 1-bit address, 4-bit data */
	if( quad_data && !quad_addr ) {
		/* Set read cmd to Quad Read */
		drcmr = (u32)QUAD_READ << 16;

		/* width: 1-bit cmd, 1-bit addr, 4-bit data */
#if (ADDRESS_BYTE_SIZE == 4)
		/* address: 32 bits */
		drenr = 0x00024f00UL;
#else /* ADDRESS_BYTE_SIZE == 3 */
		/* address: 24 bits */
		drenr = 0x00024700UL;
#endif
		/* Add extra Dummy cycles between address and data */
		dmdmcr = 0x00020000 | (g_QUAD_RD_DMY-1); /* 4 bit width, x cycles */
		drenr |= 0x00008000; /* Set Dummy Cycle Enable (DME) */
	}

	/* 1-bit address, 1-bit data */
	if( !quad_data && !quad_addr ) {
		/* Set read cmd to FAST Read */
		drcmr = (u32)FAST_READ << 16;

		/* width: 1-bit cmd, 1-bit addr, 1-bit data */
#if (ADDRESS_BYTE_SIZE == 4)
		/* address: 32 bits */
		drenr = 0x00004f00;
#else /* ADDRESS_BYTE_SIZE == 3 */
		/* address: 24 bits */
		drenr = 0x00004700;
#endif
		/* Add extra Dummy cycles between address and data */
		dmdmcr = 0x00000000 | (g_FAST_RD_DMY-1); /* 1 bit width, x cycles */
		drenr |= 0x00008000; /* Set Dummy Cycle Enable (DME) */
	}

	/* 4-bit address, 4-bit data */
	if( quad_addr ) {
		/* Set read cmd to Quad I/O */
		drcmr = (u32)QUAD_IO_READ << 16;

		/* width: 1-bit cmd, 4-bit addr, 4-bit data */
#if (ADDRESS_BYTE_SIZE == 4)
		/* address: 32 bits */
		drenr = 0x02024f00;
#else /* ADDRESS_BYTE_SIZE == 3 */
		/* address: 24 bits */
		drenr = 0x02024700;
#endif

		/* Use Option data regsiters to output 0x00 to write the
		   'mode' byte by sending OPD3 (at 4-bit) between address
		   and dummy */
		if ( g_QUAD_IO_RD_OPT ) {
			dropr = 0x00000000;
			drenr |= 0x00200080;	// send OPD3(8-bit) at 4-bit width (2 cycles total)
		}

		/* Add extra Dummy cycles between address and data */
		dmdmcr = 0x00020000 | (g_QUAD_IO_RD_DMY-1); /* 4 bit size, x cycles */

		drenr |= 0x00008000; /* Set Dummy Cycle Enable (DME) */
	}

	if ( ddr ) {
		printf( "WARNING: DDR mode doesn't actually work yet on the RSKRZA1 board.\n"
			"   The Spansion SPI flash has an extra phase in the command stream\n"
			"   that we can't account for.\n");

		/* Set read cmd to Read DDR Quad I/O */
		drcmr = (u32)QUAD_IO_DDR_READ << 16;

		/* Address, option and data all 4-bit DDR */
		drdrenr = 0x00000111;

		/* According to the Spansion spec (Table 8.5), dummy cycles
		   are needed when LC=00b for READ DDR QUAD I/O commnds */
		/* Add extra Dummy cycles between address and data */
		dmdmcr = 0x00020000 | (g_QUAD_IO_DDR_RD_DMY-1); /* 4 bit size, x cycles */
		drenr |= 0x00008000; /* Set Dummy Cycle Enable (DME) */
	}
	else {
		drdrenr = 0;
	}

	/* Set new register values */
	*(volatile u32 *)DMDMCR_0 = dmdmcr;
	*(volatile u32 *)DRENR_0 = drenr;
	*(volatile u32 *)CMNCR_0 = cmncr;
	*(volatile u32 *)DRCMR_0 = drcmr;
	*(volatile u32 *)DROPR_0 = dropr;
	*(volatile u32 *)DRDRENR_0 = drdrenr;

	/* Allow 32MB of SPI addressing (POR default is only 16MB) */
	*(volatile u32 *)DREAR_0 = 0x00000001;

	/* Turn Read Burst on, Burst Length=2 uints (also set cache flush) */
	/* Keep SSL low (SSLE=1) in case the next transfer is continugous with
	   our last...saves on address cycle. */
	*(u32 *)DRCR_0 = 0x00010301;
	asm("nop");
	*(volatile u32 *)DRCR_0;	/* Read must be done after cache flush */

	/* Do some dummy reads (our of order) to help clean things up */
	*(volatile u32 *)0x18000010;
	*(volatile int *)0x18000000;

	printf("New Mode: ");
	cmd = (*(volatile long *)DRCMR_0 >> 16) & 0xFF;
	for( i=0; i < READ_MODES; i++) {
		if( modes[i].cmd == cmd )
			printf("%s\n",modes[i].name);
	}

	return 0;
usage:
	return CMD_RET_USAGE;
}
static char qspi_help_text[] =
	"Set the XIP Mode for QSPI\n"
	"Usage: qspi [single|dual] [a1|(a4)] [d1|(d4)] [(sdr)|ddr]\n"
	"  (xx) means defualt value if not specified\n"
	"  'a4' requries 'd4' to be set\n"
	"  'ddr' requries 'd4' and 'a4' to be set\n";
U_BOOT_CMD(
	qspi,	CONFIG_SYS_MAXARGS,	1,	do_qspi,
	"Change QSPI XIP Mode", qspi_help_text
);

int do_force_updater(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	start_a03(FORCE_UPDATER);

	return 0;
}
static char force_updater_help_text[] =
	"Fource updater Start\n"
	"Usage: fouce_updater\n";

U_BOOT_CMD(
	force_updater,	CONFIG_SYS_MAXARGS,	1,	do_force_updater,
	"Fource updater Start QSPI XIP Mode", force_updater_help_text
);

int do_force_rescue(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	start_a03(FORCE_RESCUE);

	return 0;
}
static char force_rescue_help_text[] =
	"Fource rescue Start\n"
	"Usage: fouce_rescue\n";

U_BOOT_CMD(
	force_rescue,	CONFIG_SYS_MAXARGS,	1,	do_force_rescue,
	"Fource resuce Start QSPI XIP Mode", force_rescue_help_text
);

void power_led(u8 on)
{
	/* Not detected SW/LED Board */
	if (led_sw_flag != 1) {
		return;
	}

	/* P1_1 = IO_1 Power LED */
	wait();
	gpio_set(1, 1, on);
}

void wlan_led(u8 on)
{
	/* Not detected SW/LED Board */
	if (led_sw_flag != 1) {
		return;
	}

	/* P1_5 = IO_5 WLAN LED */
	wait();
	gpio_set(1, 5, on);
}

void link_led(u8 on)
{
	/* Not detected SW/LED Board */
	if (led_sw_flag != 1) {
		return;
	}

	/* P1_7 = IO_7 Link LED */
	wait();
	gpio_set(1, 7, on);
}

u8 led_sw_board(void)
{
	/* P1_8 = IO_6 SW/LED */
	wait();
	return gpio_read(1, 8);
}

u8 wlan_sw(void)
{
	/* Not detected SW/LED Board */
	if (led_sw_flag != 1) {
		return SW_OFF;
	}

	/* P1_4 = IO_4 WLAN SW */
	wait();
	return gpio_read(1, 4);
}

u8 link_sw(void)
{
	/* Not detected SW/LED Board */
	if (led_sw_flag != 1) {
		return SW_OFF;
	}

	/* P1_6 = IO_6 LINK SW */
	wait();
	return gpio_read(1, 6);
}

/* BRK信号PINに変更 */
void enable_uart_brk_pin(void) {
	/* UART TxDを GPIOに変更 */
	pfc_set_gpio(3, 15, GPIO_OUT);
	wait();
}

/* UART TxD1に変更 */
void disable_uart_brk_pin(void) {
	pfc_set_pin_function(3, 15, ALT5, 0, 0);    /* P3_15 = TxD1 */
}

/* BRK信号送信 */
void brk_send(u8 on)
{
	wait();
	gpio_set(3, 15, on);
}

#define BRK_TOGGLE_TIME		(1000)	/* ms */
void brk_toggle(void)
{
	static int brk_flag = 1;
	static ulong time_save = 0;

	if ((get_timer(0) - time_save) >= BRK_TOGGLE_TIME) {
		if (brk_flag == 1) {
			brk_send(BRK_ON);
			brk_flag = 0;
		} else {
			brk_send(BRK_OFF);
			brk_flag = 1;
		}
		time_save = get_timer(0);
	}
}

#define LED_ON_TIME			(250)	/* ms */
#define LED_OFF_TIME		(750)	/* ms */
void led_flush(void)
{
	static int led_flag = 1;
	static int brk_flag = 1;
	static ulong time_save = 0;

	if (led_flag == 1) {
		/* OFF check */
		if ((get_timer(0) - time_save) >= LED_ON_TIME) {
			wlan_led(LED_OFF);
			link_led(LED_OFF);
			power_led(LED_OFF);
			if (brk_flag == 1) {
				brk_send(BRK_ON);
				brk_flag = 0;
			} else {
				brk_send(BRK_OFF);
				brk_flag = 1;
			}
			time_save = get_timer(0);
			led_flag = 0;
		}
	} else {
		/* ON check */
		if ((get_timer(0) - time_save) >= LED_OFF_TIME) {
			wlan_led(LED_ON);
			link_led(LED_ON);
			power_led(LED_ON);
			time_save = get_timer(0);
			led_flag = 1;
		}
	}
}

/* GPIO debug */
int do_read_updater(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	u8 on_off = 0;
	u8 brk_on_off = 0;
	int i = 0;


	/* GPIOのスイッチ入力確認 */
	printf("WLAN SW %d\n", wlan_sw());
	printf("LINK SW %d\n", link_sw());

	if (argc < 3) {
		goto usage;
	}

	if (strcmp(argv[2], "on") == 0) {
		/* 不論理の為 onの場合は0を設定する */
		on_off = 0;
	} else {
		on_off = 1;
	}

	if (strcmp(argv[1], "wlan") == 0) {
		wlan_led(on_off);
	} else if (strcmp(argv[1], "link") == 0) {
		link_led(on_off);
	} else if (strcmp(argv[1], "power") == 0) {
		power_led(on_off);
	} else {
		goto usage;
	}

	/* 250msec ごとにチェック */
	while ((wlan_sw() == SW_ON) && (link_sw() == SW_ON)) {

		/* SWが押下され続けている場合はBRK信号を 500msec間隔で送出する */
		if ((i % 2) == 0) {
			brk_on_off = BRK_ON;
		} else {
			brk_on_off = BRK_OFF;
		}

		if ((i % 4) == 0) {
			on_off = LED_ON;
			i = 0;
		} else {
			on_off = LED_OFF;
		}

		wlan_led(on_off);
		link_led(on_off);
		power_led(on_off);
		brk_send(brk_on_off);
		i++;
		
		mdelay(250);
	}
	return 0;
usage:
	return CMD_RET_USAGE;
}

static char gpio_check_help_text[] =
	"gpio check\n"
	"Usage: gpio_check [led_type] [on|off]\n"
	"       [led_type] = [wlan|link|power]\n";

U_BOOT_CMD(
	gpio_check,	CONFIG_SYS_MAXARGS,	1,	do_read_updater,
	"sw led check\n", gpio_check_help_text
);
/* GPIO debug */


/* load debug */
int do_load_rescue(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret = 0;
	ret = load_rescue((u32)UPDATER_LOAD_ADDR);
	printf("load_rescue ret %d\n", ret);

	return 0;
}

static char help_text[] =
	"none\n";

U_BOOT_CMD(
	load_res,	CONFIG_SYS_MAXARGS,	1,	do_load_rescue,
	"load rescue\n", help_text
);
/* load debug */
