/*
 * Rescue for HW-A03
 *
 * Copyright (C) 2016 SHARP.Co.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

/**
 * updater のロード
 * <pre>
 *  md5の処理を行う
 * </pre>
 * @version 0.01
*/

#include "rescue.h"
#include "rom_image.h"
#include "md5.h"
#include "aes.h"

extern u8 enc_key[16];	/* 復号キー   */
extern u8 enc_iv[16];	/* 初期ベクタ */

u8 rescue_read_buffer[YBUFFER];					// データ

#define UPDATER_DEBUG
#ifdef UPDATER_DEBUG
void updater_dump(uint8_t *data, int length)
{
	char temp[80];
	int ix, iy;

#define	LP(x) (&x[strlen(x)])

	printf("data addr=%p length=%d\n", data, length);

	for (ix = 0; ix < length; ix += 16) {
		memset(temp, 0x00, 80);
		sprintf (LP(temp), "%04x  ", ix);

		for (iy = 0; iy < 16; iy ++) {
		    if (ix + iy < length) {
				unsigned int c = (unsigned int)(data[ix + iy] & 0xff);
				sprintf (LP(temp), "%02x ", c);
			} else {
				strcat (temp, "   ");
			}
	    }
		strcat (temp, "  ");

		for (iy = 0; iy < 16; iy ++) {
			if (ix + iy < length) {
				signed char c = data[ix + iy];
				if (c < ' ') {
		    		c = '.';
				}
                sprintf (LP(temp), "%c", c);
			} else {
				strcat (temp, " ");
			}
	    }
		printf ("%s\n", temp);
	}
}
#endif /* UPDATER_DEBUG */

int rescue_image_load(u8 *roffset, u8 *woffset, u32 size, u8 *md5_hash)
{
	int rc;
	int ret = -1;
	u8 *waddr;
	u32 rsize = 0;
	u32 wsize = 0;

	/* save data */
	waddr = woffset;
	wsize = size;

	/* write */
	while (wsize > 0) {
		if (wsize <= UPDATER_BLOCK_SIZE) {
			rsize = wsize;
		} else {
			rsize = UPDATER_BLOCK_SIZE;
		}
	    rc = rescue_read(roffset, rsize);
		if (rc != 0) {
			printf("rescue_read error: size=%d\n", rsize);
			goto END;
		}
		AES128_CBC_decrypt_buffer(waddr, roffset, rsize, 0, 0);
		wsize -= rsize;
		waddr += rsize;
	}

	ret = 0;
END:
	return ret;
}

/**
 * @brief アップデータのロード
 * @attention なし
 * @note 
 * @param [IN] loadaddr
 * @return OK/ERROR
 */
int load_rescue(u32 loadaddr)
{
	int ret = -1;
	int rc = -1;
	u8 *output;		/* アップデータ展開先(復号後) */
	u8 *input;		/* アップデータ暗号 */
	u8 md5_comp[MD5_HASH_SIZE];
	bool abort_flag = false;
	u16 type_code = 0;
	u32 rest_size = 0;
	u32 workaddr = loadaddr;
	rescue_file_header_t rescue_file_hdr;

	/* レスキューダウンロード待ちは クラウドLEDのみ点灯する */
	power_led(LED_OFF);
	wlan_led(LED_OFF);
	link_led(LED_ON);

	/* ymodem転送の初期化を行う */
	rc = rescue_init();
	if (rc != 0) {
		DBG_PRINT("read error: rescue image load error\n");
		goto END;
	}

	/* ROMイメージヘッダ情報を取得する */
	rc = rescue_read(rescue_read_buffer, YBUFFER);
	if (rc != 0) {
		DBG_PRINT("read error: 1st packet\n");
		goto END;
	}

	/* copy rescue image header */
	memcpy(&rescue_file_hdr, rescue_read_buffer, sizeof(rescue_file_header_t));

	/* copy rescue image in 1st packet*/
	rest_size = YBUFFER - sizeof(rescue_file_header_t);
	AES128_CBC_decrypt_buffer((u8 *)loadaddr, 
		rescue_read_buffer + sizeof(rescue_file_header_t), 
		rest_size, enc_key, enc_iv);
	workaddr += rest_size;

	type_code = be16toh(rescue_file_hdr.rom_image_hdr.type_code);

	/* check type code */
	if (type_code != RESCUE_HEADER_TYPE_CODE_RESCUE_IMAGE) {
		DBG_PRINT("error: rescue rom image header type code 0x%x\n", type_code);
		goto END;
	}

	/* check chunk type */
	if (strncmp((char*)rescue_file_hdr.rescue_image_hdr.chunk_type, 
			(char*)UPDATE_HEADER_CHUNK_RESCUE_IMAGE,
			(sizeof (rescue_file_hdr.rescue_image_hdr.chunk_type))) != 0) {
		DBG_PRINT("error: rescue image header chunk\n");
		goto END;
	}

	output = (u8 *)workaddr;
	input = rescue_read_buffer;
	rc = rescue_image_load(input, output, 
		be32toh(rescue_file_hdr.rescue_image_hdr.rescue_size) - rest_size, 
		rescue_file_hdr.rescue_image_hdr.rescue_hash);
	/* OK or NGの判断 */
	if (rc != 0) {
		DBG_PRINT("error: rescue load error\n");
		goto END;
	}

	/* check md5sum */
	updater_md5_hash(md5_comp, (u8 *)loadaddr, 
		be32toh(rescue_file_hdr.rescue_image_hdr.rescue_size));
	rc = updater_md5_hash_check(md5_comp, 
		rescue_file_hdr.rescue_image_hdr.rescue_hash);
	if (rc != UPDATER_TRUE) {
		printf("md5sum error: addr=0x%x size=%d\n", (u32)loadaddr, 
			(int)be32toh(rescue_file_hdr.rescue_image_hdr.rescue_size));
		goto END;
	}

	ret = 0;
END:
	/* ymodem転送終了処理 */
	if (ret != 0) {
		abort_flag = true;
	}
	rescue_end(abort_flag);

	return ret;
}

