/*
 * Updater for HW-A03
 *
 * Copyright (C) 2016 SHARP.Co.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "updater.h"
#include "rom_image.h"
#include "md5.h"
#include "aes.h"

extern u8 enc_key[16];	/* 復号キー   */
extern u8 enc_iv[16];	/* 初期ベクタ */

#define updater_dct_get_inactive(act_offset)	\
			((act_offset == UPDATER_DCT1_OFFSET) ? \
			UPDATER_DCT2_OFFSET: UPDATER_DCT1_OFFSET)

static u8 act_dct[UPDATER_DCT_SIZE];
static u8 inact_dct[UPDATER_DCT_SIZE];

static int rescue_sw_on = 0;

/**
 * @brief 有効DCT取得
 * @attention なし
 * @note 有効なDCTのオフセットを取得する
 * @param なし
 * @return 有効なDCTのオフセット（NULL=なし）
 */
u32 updater_dct_get_active(void)
{
	int ret;
	u32 act_offset = 0;
	dct_header_t dct1 = {{0},{0}};
	dct_header_t dct2 = {{0},{0}};

	/* read DCT1 */
	ret = updater_rom_read((u8 *)UPDATER_DCT1_OFFSET, 
			(u8 *)&dct1, sizeof (dct_header_t));
	if (ret != 0) {
		printf("read error: DCT1\n");
		goto END;
	}

	/* read DCT2 */
	ret = updater_rom_read((u8 *)UPDATER_DCT2_OFFSET, 
			(u8 *)&dct2, sizeof (dct_header_t));
	if (ret != 0) {
		printf("read error: DCT2\n");
		goto END;
	}

	if ((dct1.p_dct_header.is_current_dct == 1) &&
		(dct1.p_dct_header.write_incomplete == 0) && 
		(dct1.p_dct_header.magic_number == htole32(DCT_MAGIC))) {
		act_offset = UPDATER_DCT1_OFFSET;
	} else if ((dct2.p_dct_header.is_current_dct == 1) &&
			   (dct2.p_dct_header.write_incomplete == 0) && 
			   (dct2.p_dct_header.magic_number == htole32(DCT_MAGIC))) {
		act_offset = UPDATER_DCT2_OFFSET;
	}

END:
	return act_offset;
}

/**
 * @brief DCT headerリード処理
 * @attention なし
 * @note DCTヘッダの読み込む
 * @param [OUT] *dct_hdr: DCT header
 * @return 0: success / -1: error
 */
int updater_dct_header_read(dct_header_t *dct_hdr)
{
	int rc;
	int ret = -1;
	u32 act_offset = 0;

	/* Init QSPI ROM(SPI mode & unprotect) */
	rc = updater_rom_initial_retry();
	if (rc != 0) {
		printf("updater_rom_initial_retry error\n");
		goto END;
	}

	/* check active offset */
	act_offset = updater_dct_get_active();
	if (act_offset == 0) {
		printf("error: updater_dct_get_active\n");
		goto END;
	}

	/* read active DCT */
	rc = updater_rom_read((u8 *)act_offset, (u8 *)dct_hdr, sizeof(dct_header_t));
	if (rc != 0) {
		printf("read error: active DCT header\n");
		goto END;
	}

	/* change exmode */
	updater_rom_state(QSPI_ROM_STATE_PROTECTED);
	rc = updater_rom_deinitial_retry();
	if (rc != 0) {
		printf("updater_rom_deinitial_retry error\n");
		goto END;
	}

	ret = 0;
END:
	return ret;
}

/**
 * @brief DCT headerライト処理
 * @attention なし
 * @note DCTヘッダの書き換える
 * @param [IN] dct_hdr: Boot状態
 * @return 0: success / -1: error
 */
int updater_dct_header_write(dct_header_t *dct_hdr)
{
	int rc;
	int ret = -1;
	u32 data_size;
	u32 act_offset = 0;
	u32 inact_offset = 0;
	dct_header_t *act_dct_hdt;
	dct_header_t *inact_dct_hdt;

	/* Init QSPI ROM(SPI mode & unprotect) */
	rc = updater_rom_initial_retry();
	if (rc != 0) {
		printf("updater_rom_initial_retry error\n");
		goto END;
	}

	/* check active offset */
	act_offset = updater_dct_get_active();
	if (act_offset == 0) {
		printf("error: updater_dct_get_active\n");
		goto END;
	}
	inact_offset = updater_dct_get_inactive(act_offset);

	/* read active DCT */
	rc = updater_rom_read((u8 *)act_offset, act_dct, UPDATER_DCT_SIZE);
	if (rc != 0) {
		printf("read error: active DCT\n");
		goto END;
	}
	act_dct_hdt = (dct_header_t *)act_dct;

	/* copy data (active -> inactive) */
	memset(inact_dct, 0xff, UPDATER_DCT_SIZE);
	data_size = le32toh(act_dct_hdt->p_dct_header.used_size) - sizeof(dct_header_t);
	memcpy(inact_dct + sizeof(dct_header_t), act_dct + sizeof(dct_header_t), data_size);

	/* erase inactive DCT */
	updater_rom_erase(inact_offset, UPDATER_DCT_SIZE);
	rc = updater_rom_erase_check(inact_offset, UPDATER_DCT_SIZE);
	if (rc != 0) {
		printf("erase error: addr=%08x size=%d\n", inact_offset, UPDATER_DCT_SIZE);
		goto END;
	}

	/* write inactive DCT data */
	/* ページ境界アドレスを指定しないとエラーとなるため、	*/
	/* 0xffで初期化してヘッダを含めライト					*/
    rc = updater_rom_write_no_erase((u8 *)inact_offset, 
			(u8 *)inact_dct, le32toh(act_dct_hdt->p_dct_header.used_size));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", 
				inact_offset + sizeof(dct_header_t), le32toh(act_dct_hdt->p_dct_header.used_size));
		goto END;
	}

	/* write inactive DCT header (parameter -> inactive)*/
	memcpy(inact_dct, dct_hdr, sizeof(dct_header_t));
	inact_dct_hdt = (dct_header_t *)inact_dct;
	inact_dct_hdt->p_dct_header.write_incomplete = 1;
	inact_dct_hdt->p_dct_header.is_current_dct = 1;
	inact_dct_hdt->p_dct_header.magic_number = htole32(DCT_MAGIC);
	rc = updater_rom_write_no_erase((u8 *)inact_offset, (u8 *)inact_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", inact_offset, sizeof(dct_header_t));
		goto END;
	}

	/* update inactive DCT header (write incomplete -> write complete)*/
	inact_dct_hdt->p_dct_header.write_incomplete = 0;
    rc = updater_rom_write_no_erase((u8 *)inact_offset, (u8 *)inact_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", inact_offset, sizeof(dct_header_t));
		goto END;
	}

	/* update active DCT header (active -> inactive)*/
	/* イレースは行わず、フラグのみリセット */
	act_dct_hdt->p_dct_header.is_current_dct = 0;
    rc = updater_rom_write_no_erase((u8 *)act_offset, (u8 *)act_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", act_offset, sizeof(dct_header_t));
		goto END;
	}

	/* change exmode */
	updater_rom_state(QSPI_ROM_STATE_PROTECTED);
	rc = updater_rom_deinitial_retry();
	if (rc != 0) {
		printf("updater_rom_deinitial_retry error\n");
		goto END;
	}

	ret = 0;
END:
	return ret;
}

/**
 * @brief DCT Boot状態リード処理
 * @attention なし
 * @note DCTヘッダのBoot状態を読み込む
 * @param [OUT] *state: Boot状態
 * @return 0: success / -1: error
 */
int updater_dct_state_read(u8 *state)
{
	int rc;
	int ret = -1;
	u32 act_offset = 0;
	dct_header_t *act_dct_hdt;

	/* Init QSPI ROM(SPI mode & unprotect) */
	rc = updater_rom_initial_retry();
	if (rc != 0) {
		printf("updater_rom_initial_retry error\n");
		goto END;
	}

	/* check active offset */
	act_offset = updater_dct_get_active();
	if (act_offset == 0) {
		printf("error: updater_dct_get_active\n");
		goto END;
	}

	/* read active DCT */
	rc = updater_rom_read((u8 *)act_offset, act_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("read error: active DCT header\n");
		goto END;
	}
	act_dct_hdt = (dct_header_t *)act_dct;
	*state = act_dct_hdt->p_dct_btldt.hms_boot_state.data;

	/* change exmode */
	updater_rom_state(QSPI_ROM_STATE_PROTECTED);
	rc = updater_rom_deinitial_retry();
	if (rc != 0) {
		printf("updater_rom_deinitial_retry error\n");
		goto END;
	}

	ret = 0;
END:
	return ret;
}

/**
 * @brief DCT Boot状態ライト処理
 * @attention なし
 * @note DCTヘッダのBoot状態を書き換える
 * @param [IN] state: Boot状態
 * @return 0: success / -1: error
 */
int updater_dct_state_write(u8 *state)
{
	int rc;
	int ret = -1;
	u32 data_size;
	u32 act_offset = 0;
	u32 inact_offset = 0;
	dct_header_t *act_dct_hdt;
	dct_header_t *inact_dct_hdt;

	/* Init QSPI ROM(SPI mode & unprotect) */
	rc = updater_rom_initial_retry();
	if (rc != 0) {
		printf("updater_rom_initial_retry error\n");
		goto END;
	}

	/* check active offset */
	act_offset = updater_dct_get_active();
	if (act_offset == 0) {
		printf("error: updater_dct_get_active\n");
		goto END;
	}
	inact_offset = updater_dct_get_inactive(act_offset);

	/* read active DCT */
	rc = updater_rom_read((u8 *)act_offset, act_dct, UPDATER_DCT_SIZE);
	if (rc != 0) {
		printf("read error: active DCT\n");
		goto END;
	}
	act_dct_hdt = (dct_header_t *)act_dct;

	/* copy data (active -> inactive) */
	memset(inact_dct, 0xff, UPDATER_DCT_SIZE);
	data_size = le32toh(act_dct_hdt->p_dct_header.used_size) - sizeof(dct_header_t);
	memcpy(inact_dct + sizeof(dct_header_t), act_dct + sizeof(dct_header_t), data_size);

	/* erase inactive DCT */
	updater_rom_erase(inact_offset, UPDATER_DCT_SIZE);
	rc = updater_rom_erase_check(inact_offset, UPDATER_DCT_SIZE);
	if (rc != 0) {
		printf("erase error: addr=%08x size=%d\n", inact_offset, UPDATER_DCT_SIZE);
		goto END;
	}

	/* write inactive DCT data */
	/* ページ境界アドレスを指定しないとエラーとなるため、	*/
	/* 0xffで初期化してヘッダを含めライト					*/
    rc = updater_rom_write_no_erase((u8 *)inact_offset, 
			(u8 *)inact_dct, le32toh(act_dct_hdt->p_dct_header.used_size));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", 
				inact_offset + sizeof(dct_header_t), le32toh(act_dct_hdt->p_dct_header.used_size));
		goto END;
	}

	/* write inactive DCT header (active -> inactive)*/
	memcpy(inact_dct, act_dct, sizeof(dct_header_t));
	inact_dct_hdt = (dct_header_t *)inact_dct;
	inact_dct_hdt->p_dct_header.write_incomplete = 1;
	inact_dct_hdt->p_dct_header.is_current_dct = 1;
	inact_dct_hdt->p_dct_header.magic_number = htole32(DCT_MAGIC);
	inact_dct_hdt->p_dct_btldt.hms_boot_state.data = *state;
	rc = updater_rom_write_no_erase((u8 *)inact_offset, (u8 *)inact_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", inact_offset, sizeof(dct_header_t));
		goto END;
	}

	/* update inactive DCT header (write incomplete -> write complete)*/
	inact_dct_hdt->p_dct_header.write_incomplete = 0;
    rc = updater_rom_write_no_erase((u8 *)inact_offset, (u8 *)inact_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", inact_offset, sizeof(dct_header_t));
		goto END;
	}

	/* update active DCT header (active -> inactive)*/
	/* イレースは行わず、フラグのみリセット */
	act_dct_hdt->p_dct_header.is_current_dct = 0;
    rc = updater_rom_write_no_erase((u8 *)act_offset, (u8 *)act_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", act_offset, sizeof(dct_header_t));
		goto END;
	}

	/* change exmode */
	updater_rom_state(QSPI_ROM_STATE_PROTECTED);
	rc = updater_rom_deinitial_retry();
	if (rc != 0) {
		printf("updater_rom_deinitial_retry error\n");
		goto END;
	}

	ret = 0;
END:
	return ret;
}

/**
 * @brief DCT リトライカウントリード処理
 * @attention なし
 * @note DCTヘッダのリトライカウントを読み込み
 * @param [OUT] *retry: リトライカウント
 * @return 0: success / -1: error
 */
int updater_dct_retry_read(u8 *retry)
{
	int rc;
	int ret = -1;
	u32 act_offset = 0;
	dct_header_t *act_dct_hdt;

	/* Init QSPI ROM(SPI mode & unprotect) */
	rc = updater_rom_initial_retry();
	if (rc != 0) {
		printf("updater_rom_initial_retry error\n");
		goto END;
	}

	/* check active offset */
	act_offset = updater_dct_get_active();
	if (act_offset == 0) {
		printf("error: updater_dct_get_active\n");
		goto END;
	}

	/* read active DCT */
	rc = updater_rom_read((u8 *)act_offset, act_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("read error: active DCT header\n");
		goto END;
	}
	act_dct_hdt = (dct_header_t *)act_dct;
	*retry = act_dct_hdt->p_dct_btldt.hms_update_retry_cnt;

	/* change exmode */
	updater_rom_state(QSPI_ROM_STATE_PROTECTED);
	rc = updater_rom_deinitial_retry();
	if (rc != 0) {
		printf("updater_rom_deinitial_retry error\n");
		goto END;
	}

	ret = 0;
END:
	return ret;
}


/**
 * @brief DCT リトライカウントライト処理
 * @attention なし
 * @note DCTヘッダのリトライカウントを書き換える
 * @param [IN] retry: リトライカウント
 * @return 0: success / -1: error
 */
int updater_dct_retry_write(u8 *retry)
{
	int rc;
	int ret = -1;
	u32 data_size;
	u32 act_offset = 0;
	u32 inact_offset = 0;
	dct_header_t *act_dct_hdt;
	dct_header_t *inact_dct_hdt;

	/* Init QSPI ROM(SPI mode & unprotect) */
	rc = updater_rom_initial_retry();
	if (rc != 0) {
		printf("updater_rom_initial_retry error\n");
		goto END;
	}

	/* check active offset */
	act_offset = updater_dct_get_active();
	if (act_offset == 0) {
		printf("error: updater_dct_get_active\n");
		goto END;
	}
	inact_offset = updater_dct_get_inactive(act_offset);

	/* read active DCT */
	rc = updater_rom_read((u8 *)act_offset, act_dct, UPDATER_DCT_SIZE);
	if (rc != 0) {
		printf("read error: active DCT\n");
		goto END;
	}
	act_dct_hdt = (dct_header_t *)act_dct;

	/* copy data (active -> inactive) */
	memset(inact_dct, 0xff, UPDATER_DCT_SIZE);
	data_size = le32toh(act_dct_hdt->p_dct_header.used_size) - sizeof(dct_header_t);
	memcpy(inact_dct + sizeof(dct_header_t), act_dct + sizeof(dct_header_t), data_size);

	/* erase inactive DCT */
	updater_rom_erase(inact_offset, UPDATER_DCT_SIZE);
	rc = updater_rom_erase_check(inact_offset, UPDATER_DCT_SIZE);
	if (rc != 0) {
		printf("erase error: addr=%08x size=%d\n", inact_offset, UPDATER_DCT_SIZE);
		goto END;
	}

	/* write inactive DCT data */
	/* ページ境界アドレスを指定しないとエラーとなるため、	*/
	/* 0xffで初期化してヘッダを含めライト					*/
	rc = updater_rom_write_no_erase((u8 *)inact_offset, 
			(u8 *)inact_dct, le32toh(act_dct_hdt->p_dct_header.used_size));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", 
				inact_offset + sizeof(dct_header_t), le32toh(act_dct_hdt->p_dct_header.used_size));
		goto END;
	}

	/* write inactive DCT header (active -> inactive)*/
	memcpy(inact_dct, act_dct, sizeof(dct_header_t));
	inact_dct_hdt = (dct_header_t *)inact_dct;
	inact_dct_hdt->p_dct_header.write_incomplete = 1;
	inact_dct_hdt->p_dct_header.is_current_dct = 1;
	inact_dct_hdt->p_dct_header.magic_number = htole32(DCT_MAGIC);
	inact_dct_hdt->p_dct_btldt.hms_update_retry_cnt = *retry;
	rc = updater_rom_write_no_erase((u8 *)inact_offset, (u8 *)inact_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", inact_offset, sizeof(dct_header_t));
		goto END;
	}

	/* update inactive DCT header (write incomplete -> write complete)*/
	inact_dct_hdt->p_dct_header.write_incomplete = 0;
    rc = updater_rom_write_no_erase((u8 *)inact_offset, (u8 *)inact_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", inact_offset, sizeof(dct_header_t));
		goto END;
	}

	/* update active DCT header (active -> inactive)*/
	/* イレースは行わず、フラグのみリセット */
	act_dct_hdt->p_dct_header.is_current_dct = 0;
    rc = updater_rom_write_no_erase((u8 *)act_offset, (u8 *)act_dct, sizeof(dct_header_t));
	if (rc != 0) {
		printf("write error: addr=%08x size=%d\n", act_offset, sizeof(dct_header_t));
		goto END;
	}

	/* change exmode */
	updater_rom_state(QSPI_ROM_STATE_PROTECTED);
	rc = updater_rom_deinitial_retry();
	if (rc != 0) {
		printf("updater_rom_deinitial_retry error\n");
		goto END;
	}

	ret = 0;
END:
	return ret;
}


int updater_image_write(u8 *woffset, u32 size, u8 *md5_hash)
{
	int rc;
	int ret = -1;
	int first_flag = UPDATER_FALSE;
	u8 md5_comp[MD5_HASH_SIZE];
	u8 *waddr;
	u32 rsize = 0;
	u32 wsize = 0;

	DBG_PRINT("roffset  = %p\n", roffset);
	DBG_PRINT("woffset  = %p\n", woffset);
	DBG_PRINT("size     = %d\n", size);

	/* Init QSPI ROM(SPI mode & unprotect) */
	rc = updater_rom_initial_retry();
	if (rc != 0) {
		printf("updater_rom_initial_retry error\n");
		goto END;
	}

	/* erase */
	DBG_PRINT("updater_rom_erase() start\n");
	updater_rom_erase((u32)woffset, size);
	DBG_PRINT("updater_rom_erase() end\n");

	DBG_PRINT("updater_rom_erase_check() start\n");
	rc = updater_rom_erase_check((u32)woffset, size);
	if (rc != 0) {
		printf("erase error: addr=%x size=%d\n", (u32)woffset, size);
		goto END;
	}
	DBG_PRINT("updater_rom_erase_check() end\n");

	/* write */
	DBG_PRINT("write start\n");
	waddr = woffset;
	wsize = size;

	while (wsize > 0) {
		if (wsize <= UPDATER_BLOCK_SIZE) {
			rsize = wsize;
		} else {
			rsize = UPDATER_BLOCK_SIZE;
		}
	    rc = updater_ffs_read(updater_read_buffer, rsize);
		if (rc != 0) {
			printf("updater_ffs_read error: size=%d\n", rsize);
			goto END;
		}

		if (first_flag == UPDATER_FALSE) {
			AES128_CBC_decrypt_buffer(updater_write_buffer, updater_read_buffer, 
										rsize, enc_key, enc_iv);
			first_flag = UPDATER_TRUE;
		} else {
			AES128_CBC_decrypt_buffer(updater_write_buffer, updater_read_buffer, 
										rsize, 0, 0);
		}
	    rc = updater_rom_write_no_erase((u8 *)waddr, updater_write_buffer, rsize);
		if (rc != 0) {
			printf("write error: addr=%x size=%d\n", (u32)waddr, rsize);
			goto END;
		}

		waddr += rsize;
		wsize -= rsize;
		printf("%02d%% write\n", ((size - wsize)*100) / size);
	}
	printf("\n");
	DBG_PRINT("write end\n");

	/* change exmode */
	updater_rom_state(QSPI_ROM_STATE_PROTECTED);
	rc = updater_rom_deinitial_retry();
	if (rc != 0) {
		printf("updater_rom_deinitial_retry error\n");
		goto END;
	}

	/* check md5sum */
	DBG_PRINT("check md5sum start\n");
	updater_md5_hash(md5_comp, woffset, size);
	rc = updater_md5_hash_check(md5_comp, md5_hash);
	if (rc != UPDATER_TRUE) {
		printf("md5sum error: addr=%x size=%d\n", (u32)woffset, size);
		goto END;
	}
	DBG_PRINT("check md5sum end\n");

	ret = 0;
END:
	return ret;
}

int rescue_mode_sw(void)
{
	u8 on_off = 0;
	u8 brk_on_off = 0;
	int i = 0;
	int ret = 0;

	while ((wlan_sw() == SW_ON) && (link_sw() == SW_ON)) {
		ret = 1;

		/* SWが押下され続けている場合はBRK信号を 500msec間隔で送出する */
		if ((i % 2) == 0) {
			brk_on_off = BRK_ON;
		} else {
			brk_on_off = BRK_OFF;
		}

		if ((i % 4) == 0) {
			on_off = LED_ON;
			i = 0;
		} else {
			on_off = LED_OFF;
		}

		wlan_led(on_off);
		link_led(on_off);
		power_led(on_off);
		brk_send(brk_on_off);
		i++;

		/* 250msec ごとにチェック */
		mdelay(250);
	}
	return ret;
}


/* DCTのbootstateによらず u-bootからレスキュー起動/アップデータ起動モード
 * になる場合は DCTが2面ともに未初期化の場合に初期DCTにて初期化する 
 */
int updater_initial_dct(void)
{
	int rc;
	int ret = -1;
	linux_image_header_t linux_img_hdr;
	u32 size = 0;
	u32 rsize = 0;
	dct_header_t *dct_offset = NULL;
	u8 *waddr = 0;
	u8 *raddr = 0;

	/* Flashの初期設定を実施 */
	rc = updater_rom_initial_retry();

	if (rc != 0) {
		printf("flash error: flash init error\n");
		goto END;
	}

	/* 有効なDCTがあるかチェックする  */
	dct_offset = (dct_header_t*)updater_dct_get_active();
	if (dct_offset != 0) {
		/* 有効なDCTがある場合は初期DCTの確認を行わない */
		ret = 0;
		goto END;
	}

    rc = updater_ffs_read_init(UPDATER_FIRM_OFFSET);
	if (rc != 0) {
		printf("updater_ffs_read_init error: addr=%x\n", (u32)UPDATER_FIRM_OFFSET);
		goto END;
	}
	rc = updater_ffs_read((u8 *)&linux_img_hdr, sizeof(linux_image_header_t));
	if (rc != 0) {
		printf("updater_ffs_read error\n");
		goto END;
	}

	/* FFS形式で読み込む為、初期DCTが存在する部分まで、読み飛ばす */
	size = be32toh(linux_img_hdr.updater_size) + be32toh(linux_img_hdr.devicetree_size)+
		   be32toh(linux_img_hdr.kernel_size)  + be32toh(linux_img_hdr.rfs_size) +
		   be32toh(linux_img_hdr.exsystem_size);
	while (size > 0) {
		if (size <= UPDATER_BLOCK_SIZE) {
			rsize = size;
		} else {
			rsize = UPDATER_BLOCK_SIZE;
		}
	    rc = updater_ffs_read(updater_read_buffer, rsize);
		if (rc != 0) {
			printf("updater_ffs_read error: size=%d\n", rsize);
			goto END;
		}
		size -= rsize;
	}

	/* init DCT1/DCT2 */
	printf("init DCT1/DCT2 \n");
	size = be32toh(linux_img_hdr.dct_size);
	rc = updater_image_write((u8 *)UPDATER_DCT1_OFFSET, size, linux_img_hdr.dct_hash);
	if (rc != 0) {
		printf("error: DCT1 init\n");
		goto END;
	}

	/* DCT1の書き込み成功の場合は、書き込んだDCT1の内容をDCT2にコピーする */
	waddr = (u8*)UPDATER_DCT2_OFFSET;
	raddr = (u8*)UPDATER_DCT1_OFFSET;

	rc = updater_rom_erase((u32)waddr, size);
	if (rc != 0) {
		printf("updater_erase error: addr=0x%x\n", (u32)waddr);
		goto END;
	}

	while (size > 0) {
		if (size <= UPDATER_BLOCK_SIZE) {
			rsize = size;
		} else {
			rsize = UPDATER_BLOCK_SIZE;
		}

		rc = updater_rom_read(raddr, updater_read_buffer, rsize);
		if (rc != 0) {
			printf("updater_rom_read error: size=%d\n", rsize);
			break;
		}

		rc = updater_rom_write_no_erase(waddr, updater_read_buffer, rsize);
		if (rc != 0) {
			printf("updater_rom_write error: waddr=0x%x size=%d\n", (u32)waddr, rsize);
			break;
		}

		size  -= rsize;
		waddr += rsize;
		raddr += rsize;
	}

	if (rc != 0) {
		printf("error: DCT2 init\n");
		goto END;
	}

	ret = 0;
END:
	return ret;
}

int is_rescue_sw(void)
{
	return rescue_sw_on;
}

/**
 * @brief 起動モードの判断
 * @attention なし
 * @note 
 * @return 通常起動/レスキュー起動(Haltモード)/レスキュー起動/アップデータ起動
 */
START_MODE get_startmode(void)
{
	START_MODE mode = NORMAL_MODE;
	dct_header_t *dct_hdt = NULL;
	int ret = 0;

	/* 無線SW LINKSW 双方を押下されている場合は レスキュー起動と判断する */
	rescue_sw_on = rescue_mode_sw();

	/* DCTヘッダを取得し bootstateフラグの情報を確認 */
	dct_hdt = (dct_header_t*)updater_dct_get_active();
	if (dct_hdt == 0) {
		printf("error: DCT read error\n");

		if (rescue_sw_on) {
			/* bootstateが取得できない場合で、レスキュー起動の場合は */
			/* 初期DCTで DCT1面/2面の初期化を試みる */
			ret = updater_initial_dct();
			if (ret != 0) {
				mode = RESET_MODE;
			}
			mode = RESCUE_MODE;
		}
		goto END;
	}

	if (rescue_sw_on) {
		mode = RESCUE_MODE;
		goto END;
	}

	DBG_PRINT(" state 0x%x\n", dct_hdt->p_dct_btldt.hms_boot_state.data);

	/* 優先度はHalt→Rescue→Updater */
	if (dct_hdt->p_dct_btldt.hms_boot_state.data ==  0xFF) {
		/* 通常起動 オールFFは通常起動と判断する */
		 mode = NORMAL_MODE;
	} else if (dct_hdt->p_dct_btldt.hms_boot_state.bit.halt) {
		/* Halt */
		 mode = RESCUE_HALT_MODE;
	} else if (dct_hdt->p_dct_btldt.hms_boot_state.bit.rescue) {
		/* レスキュー起動 */
		 mode = RESCUE_MODE;
	} else if (dct_hdt->p_dct_btldt.hms_boot_state.bit.update) {
		/* アップデータ起動 */
		 mode = UPDATER_MODE;
	} else {
		/* 通常起動 */
		 mode = NORMAL_MODE;
	}

END:
	return mode;
}

void set_startmode(START_MODE mode)
{
	hms_boot_state_t set = {0};
	hms_boot_state_t set_pre = {0};

	updater_dct_state_read(&(set.data));
	set_pre = set;
	switch(mode)
	{
		case RESCUE_HALT_MODE:
			set.bit.halt = 1;
		break;
		case RESCUE_MODE:
			set.bit.rescue = 1;
		break;
		case UPDATER_MODE:
			set.bit.update = 1;
		break;
		case NORMAL_MODE:
			set.data = 0;
		break;
		case HALT_FLAG_RESET:
			set.bit.halt = 0;
		break;
		case RESCUE_FLAG_RESET:
			set.bit.rescue = 0;
		break;
		case UPDATE_FLAG_RESET:
			set.bit.update = 0;
		break;
		default:
			/* 無処理 */
		break;
	}
	if (set.data != set_pre.data) {
		updater_dct_state_write(&(set.data));
	}
}
