/*
 * Updater for HW-A03
 *
 * Copyright (C) 2016 SHARP.Co.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */


#include "updater.h"

typedef struct ffs_read_info {
	u8 *start_offset;
	u8 *end_offset;
	u8 *read_start;
	u8 *read_end;
	u32 data_offset;
	u8 data[FFS_BLOCK_SIZE];
} ffs_read_info_t;

static ffs_read_info_t ffs_read_info = {
	.start_offset = (u8 *)UPDATER_FIRM_OFFSET,
	.end_offset   = (u8 *)(UPDATER_FIRM_OFFSET + UPDATER_FIRM_SIZE)
};

/**
 * @brief FFSリード初期化処理
 * @attention なし
 * @note FFSリード処理の初期化
 * @param [IN] *start: read start offset
 * @return 0: success / -1: error
 */
int updater_ffs_read_init(u32 read_start)
{
	int ret = -1;
	int rc = -1;

	/* check parameter */
	if (read_start < (u32)ffs_read_info.start_offset || 
		read_start >= (u32)ffs_read_info.end_offset) {
	 	printf("updater_ffs_read_init: invalid parameter read_start=0x%x\n", 
				read_start);
		goto END;
	}

	/* calc read_start */
	read_start = (read_start & ~(sizeof(ffs_block_t) - 1));
	ffs_read_info.read_start = (u8 *)read_start;
	ffs_read_info.read_end = ffs_read_info.read_start;
	ffs_read_info.data_offset = 0;
	rc = updater_ffs_read_block(ffs_read_info.data, 1);
	if (rc != 0) {
	 	printf("updater_ffs_read_init: block read error\n");
		goto END;
	}

	ret = 0;
END:
	return ret;
}

/**
 * @brief FFSブロックリード処理
 * @attention なし
 * @note FFSデータのブロック読み込む
 * @param [OUT] *data: read data
 * @return 0: success / -1: error
 */
int updater_ffs_read_block(u8 *data, int top)
{
	int ret = -1;
	int rc = -1;
	int count = UPDATER_FIRM_SIZE / sizeof (ffs_block_t) + 1;
	u8 *addr;
	ffs_header_t ffs_header;

	while (count--) {
		/* read ffs header */
		rc = updater_rom_read(ffs_read_info.read_start, (u8 *)&ffs_header, sizeof(ffs_header_t));
		if (rc != 0) {
			printf("read error: ffs header addr=%p\n", ffs_read_info.read_start);
			goto END;
		}

		/* check header */
		if (top) {
			if ((ffs_header.state == FFS_STATE_USED) && 
				(ffs_header.code  == FFS_CODE_LINUX)) {
				break;
			}
		} else {
			if (ffs_header.state == FFS_STATE_USED) {
				break;
			}
		}

		/* update read_start */
		ffs_read_info.read_start += sizeof (ffs_block_t);
		if (ffs_read_info.read_start >= ffs_read_info.end_offset) {
			ffs_read_info.read_start = ffs_read_info.start_offset;
		}

		/* check rollover */
		if (ffs_read_info.read_start == ffs_read_info.read_end) {
			printf("error: ffs rollover addr=%p\n", ffs_read_info.read_start);
			goto END;
		}
	}

	if (count == 0) {
		printf("error: ffs read addr=%p\n", ffs_read_info.read_start);
		goto END;
	}

	/* read data */
	addr = ffs_read_info.read_start + sizeof (ffs_block_t) - FFS_BLOCK_SIZE;
	rc = updater_rom_read(addr, data, FFS_BLOCK_SIZE);
	if (rc != 0) {
		printf("read error: ffs data addr=%p\n", addr);
		goto END;
	}

	/* update read_start */
	ffs_read_info.read_start += sizeof (ffs_block_t);
	if (ffs_read_info.read_start >= ffs_read_info.end_offset) {
		ffs_read_info.read_start = ffs_read_info.start_offset;
	}

	ret = 0;
END:
	return ret;
}

/**
 * @brief FFSブロックリード処理
 * @attention なし
 * @note FFSデータのブロック読み込む
 * @param [OUT] *data: read data
 * @return 0: success / -1: error
 */
int updater_ffs_read(u8 *data, u32 size)
{
	int rc;
	int ret = -1;
	u32 rest_size = 0;

	if (size > FFS_BLOCK_SIZE) {
		printf("updater_ffs_read: invalid parameter(should be <=%d) size=%d\n", 
				size, FFS_BLOCK_SIZE);
		goto END;
	}

	if (size <= (FFS_BLOCK_SIZE - ffs_read_info.data_offset)) {
		memcpy(data, &(ffs_read_info.data[ffs_read_info.data_offset]), size);
		ffs_read_info.data_offset += size;
	} else {
		if (ffs_read_info.data_offset != FFS_BLOCK_SIZE) {
			rest_size = FFS_BLOCK_SIZE - ffs_read_info.data_offset;
			memcpy(data, &(ffs_read_info.data[ffs_read_info.data_offset]), rest_size);
		}
		rc = updater_ffs_read_block(ffs_read_info.data, 0);
		if (rc != 0) {
			printf("updater_ffs_read: error updater_ffs_read_block)\n");
			goto END;
		}
		memcpy(data + rest_size, ffs_read_info.data, size - rest_size);
		ffs_read_info.data_offset = size - rest_size;
	}

	ret = 0;
END:
	return ret;
}
